#******************************************************************************
#				Package For Color Definition
#******************************************************************************
# Description:
#	Procedures for color definitions, like ...
#		o Tint and mix colors	--- Okay
#		o Translate color formats, default output is #HHeexx	--- Okay
#		o Search nearest name of given color	--- Okay
#		o Takes care of additional 'SystemColors' on windows.	--- Okay
# ToDo:
#	o Create color sets, like analog, complementary, etc. from base color.
#	o Create colored states (active, selected, disabled ...) based of color
#		sets.
#
# Written by Juergen "Arthur" Beyersdorff, Stuttgart, Germany, 2013
# Rewritten by Juergen "Arthur" Beyersdorff, Stuttgart, Germany, 2016
#



package provide defColor 0.3
namespace eval ::defColor {
	variable ColorNames
	set CFId [open [file join [file dirname [info script]] rgb.txt] r]
	while {[gets $CFId Line] >= 0} {
		set SplitLine [split $Line "\t"]
		if {[string match {* *} [lindex $SplitLine 2]] > 0} {continue}
		foreach {R G B} [split [string trim [lindex [regsub -all {  *} $SplitLine { }] 0]] " "] {
		set List [list $R $G $B [lindex $SplitLine 2] [expr $R + $G + $B]]
		lappend ColorNames $List
		}
	}
}



#==============================================================================
#				Procedures
#==============================================================================

#------------------------------------------------------------------------------
#			Reverse Colors
#------------------------------------------------------------------------------
# Change order of the two given colors

proc ::defColor::reverse {Color1 Color2} {
	return [list $Color2 $Color1]
}

#------------------------------------------------------------------------------
#			Mix Colors
#------------------------------------------------------------------------------
# Mix first color with second color
# Input:	<BaseColor>
#			<MixColor> (Optional)
#			<MixValue> (Optional)
# Output:	<Color> (Hex)

proc ::defColor::mix {BaseColor {MixColor ""} {MixValue 50}} {
#	set BaseColor	[conv2List $BaseColor]
	if {$MixColor == ""} {
		set MixColor	[crFore $BaseColor]
	}
	foreach {BR BG BB} [conv2List $BaseColor] {}
	foreach {MR MG MB} [conv2List $MixColor] {}
	set R	[expr round($BR + ($MR - $BR) / (100.0 / $MixValue))]
	set G	[expr round($BG + ($MG - $BG) / (100.0 / $MixValue))]
	set B	[expr round($BB + ($MB - $BB) / (100.0 / $MixValue))]
	return [list2Hex [list $R $G $B]]
}

#------------------------------------------------------------------------------
#			Tint Color
#------------------------------------------------------------------------------
# Create lighter and darker color from given color
# Input:	<BaseColor>
#			<TintValue> (Optional)
# Output:	[list <LightColor> <DarkColor>]

proc ::defColor::tint {Color {TintValue 50}} {
	set L	[::defColor::mix $Color white $TintValue]
	set D	[::defColor::mix $Color black $TintValue]
	if {[::defColor::luminance $Color] > 90} {
		set L	[::defColor::mix $Color black [expr $TintValue / 3]]
	} elseif {[::defColor::luminance $Color] > 80} {
		set L	"#ffffff"
	}
	if {[::defColor::luminance $Color] < 15} {
		set D	[::defColor::mix $Color white [expr $TintValue / 3]]
	} elseif {[::defColor::luminance $Color] < 30} {
		set D	"#000000"
	}
	return [list $L $D]
}

#------------------------------------------------------------------------------
#			Set Foreground
#------------------------------------------------------------------------------
# Set foreground to black or white.
# Input:	<BackColor>
# Output:	<ForeColor> (Hex)

proc ::defColor::crFore {Color} {
	if {[luminance $Color] > 50} {
		return "#000000"
	} else {
		return "#ffffff"
	}
}


#------------------------------------------------------------------------------
#			Luminance
#------------------------------------------------------------------------------
# Calculate brightness
# Input:	<Color>
# Output:	<Percent>

proc ::defColor::luminance {Color} {
	foreach {r g b} [conv2List $Color] {
		set rl	[expr $r * 0.299]
		set gl	[expr $g * 0.587]
		set bl	[expr $b * 0.114]
	}
	return [expr round(($rl + $gl + $bl) * 100 / 255)]
}

#------------------------------------------------------------------------------
#			Nearest color name
#------------------------------------------------------------------------------
# Search the nearest color name
# Input: <Color>
# Output: [list <Name> <Differenz>]

proc ::defColor::nearestName {Color} {
	variable ColorNames
	set NearestDiff	[expr 255 * 3]
	foreach {r g b} [conv2List $Color] {}
	foreach C $ColorNames {
		set CR [lindex $C 0]
		set CG [lindex $C 1]
		set CB [lindex $C 2]
		set Diff	[expr abs($r - $CR) + abs($g - $CG) + abs($b - $CB)]
		if {$Diff < $NearestDiff && [catch {winfo rgb . [lindex $C 3]}] == 0} {
			set NearestDiff	$Diff
			set NearestName	[lindex $C 3]
		}
		if {$NearestDiff == 0} {
			break
		}
	}
	return [list $NearestName $NearestDiff]
}

#------------------------------------------------------------------------------
#			Helpers
#------------------------------------------------------------------------------

#---	Convert To 8bit RGB List	------------------------
# Convert color to rgb list
# Input:	<Color>
# Output:	[list R G B]

proc ::defColor::conv2List {Color} {
	if {[catch {set rgb [winfo rgb . $Color]}]} {
#		puts "Handle manually"
		if {[regexp -- {([.0-9]+)[%]?[ ]*,[ ]*([.0-9]+)[%]?[ ]*,[ ]*([.0-9]+)[%]?} $Color -> r g b]} {
#			puts "Perc or float comma separated list"
			foreach v [list $r $g $b] {
				if {[string is integer $v] == 0} {
					set r	[expr $r * 100]
					set g	[expr $g * 100]
					set b	[expr $b * 100]
					break
				}
			}
			set factor	[expr 255.0 / 100]
			set R	[expr round($r * $factor)]
			set G	[expr round($g * $factor)]
			set B	[expr round($b * $factor)]
			return [list $R $G $B]
		} elseif {[regexp -- {^([.0-9]+)[ ]+([.0-9]+)[ ]+([.0-9]+)$} $Color -> R G B]} {
#			puts "Tcl list"
			foreach v [list $R $G $B] {
				if {$v >= 256} {
					set R	[expr round($R / 256)]
					set G	[expr round($G / 256)]
					set B	[expr round($B / 256)]
					break
				}
			}
			return [list $R $G $B]
		}
	} else {
#		puts "Handle winfo = $rgb"
		foreach {r g b} $rgb {
			set R	[expr round($r / 256)]
			set G	[expr round($g / 256)]
			set B	[expr round($b / 256)]
		}
		return [list $R $G $B]
	}
	return 1
}

#---	Convert To 8bit Hex	--------------------------------
# Convert color to 8bit hex color
# Input:	[list R G B]
# Output:	#RRGGBB

proc ::defColor::conv2Hex {Color} {
	set rgb	[conv2List $Color]
#	puts "RGB List : $rgb"
	return [list2Hex $rgb]
}

# Convert rgb list to hex

proc ::defColor::list2Hex {rgb} {
	foreach {r g b} $rgb {
		set R	[format %02x $r / 256)]
		set G	[format %02x $g / 256)]
		set B	[format %02x $b / 256)]
	}
	return "#${R}${G}${B}"
}

#---	Convert To Percent List	----------------------------
# Normally not used. Only info purpose.

proc ::defColor::conv2Perc {Color} {
	set rgb	[conv2List $Color]
#	puts "RGB List : $rgb"
	foreach {r g b} $rgb {
		set R	[expr round($r * 100.0 / 255)]
		set G	[expr round($g * 100.0 / 255)]
		set B	[expr round($b * 100.0 / 255)]
	}
	return "$R,$G,$B"
}

#---	Color 2 Int	------------------------------
# Used by KomodoEdit

proc ::defColor::c2Int {Color} {
	set Dec	[conv2List $Color]
	set IntR	[lindex $Dec 0]
	set IntG	[expr [lindex $Dec 1] * 256]
	set IntB	[expr [lindex $Dec 2] * 65536]
	set Int	[expr $IntR + $IntG + $IntB]
	return $Int
}

#------------------------------------------------------------------------------
#			Compatibility
#------------------------------------------------------------------------------

proc ::defColor::isBright {Color} {
	if {[luminance $Color] > 50} {
		return 1
	} else {
		return 0
	}
}

proc ::defColor::brightLevel {Color} {
	return [luminance $Color]
}

# ---	List Names of rgb.txt	---------------------------

proc ::defColor::listNames {} {
	variable ColorNames
	foreach Item $ColorNames {
		lappend NameList [lindex $Item 3]
	}
	return [lsort -dictionary $NameList]
}

#---	Validate Color	------------------------------------
# Not needed anymore !?

proc ::defColor::validate {Color} {
	if {[regexp -- {^#[0-9A-Fa-f]*$} $Color]} {
		return HexValue
	}
	if {[regexp -- {^[0-9]{1,5}[ ]+[0-9]{1,5}[ ]+[0-9]{1,5}} $Color]} {
		return RgbList
	}
	if {[regexp -- {^[01]?.[0-9]+[ ]*,[ ]*[01]?.[0-9]+[ ]*,[ ]*[01]?.[0-9]+} $Color]} {
		return FloatValue
	}
	if {[regexp -- {^[0-9]+%[ ]*,[ ]*[0-9]+%[ ]*,[ ]*[0-9]+%} $Color]} {
		return PercValue
	}
	return 0
}

#---	Conv To Hex	----------------------------------------

proc ::defColor::c2Hex {Color} {
	return [conv2Hex $Color]
}

#------------------------------------------------------------------------------
#			Help Message
#------------------------------------------------------------------------------
# Prints help message

proc ::defColor::help {} {
	puts {
#******************************************************************************
#				::defColor
#******************************************************************************

package require defColor ?0.3?

DESCRIPTION
	Procedures for color definitions, like ...
		o Tint and mix colors
		o Translate color formats and names, default output is #HHeexx
		o Search nearest name of given color
		o Takes care of additional 'SystemColors' on windows

COMMANDS

	::defColor::reverse <Color1> <Color2>
		Reverse Colors
		Output: [list <Color2> <Color1>]

	::defColor::mix <BaseColor> ?<MixColor>? ?<MixValue>?
		Mix <BaseColor> with <MixColor> with the strength of
		<MixValue> (0-100).
		Defaults: <MixColor> defaults to black or white depending of
		Brightness of <BaseColor>. <MixValue> defaults to 50.
		Output: <HexColor>

	::defColor::tint <Color> ?<TintValue>?
		Create lighter and darker color from given color with the
		strength of <TintValue> (0-100)
		Defaults: <TintValue> defaults to 50.
		Output: [list <LighterColor> <DarkerColor>]

	::defColor::crFore <BackColor>
		Set black or white <ForeColor> depending of <BackColor>
		Output: <ForeColor>

	::defColor::luminance <Color>
		Calculate brightness level.
		Output: <Percent>

	::defColor::nearestName <Color>
		Search nearest <ColorName> of <Color>
		Output: [list <ColorName> <Differenz>]

	Internal Helpers:
		::defColor::conv2List <Color>
		::defColor::conv2Hex <Color>
		::defColor::list2Hex [list R G B]
		::defColor::conv2Perc <Color>

	Old, for compatibility:
		::defColor::isBright <Color>
		::defColor::brightLevel <Color>
		::defColor::listNames
		::defColor::validate <Color>

VARIABLES

	::defColor::ColorNames
		List of all known color names. Each color definition is a
		sublist like [list <Red> <Green> <Blue> <Name> <RGB-Sum>]
	}
}
